# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common Functions applicable across multiple components of pywbemcli
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import re
from collections import OrderedDict
import six
import click

import pywbem
from pywbem import WBEMServer, configure_loggers_from_string

from . import mockscripts
from .config import DEFAULT_URL_SCHEME, DEFAULT_CONNECTION_TIMEOUT, \
    DEFAULT_NAMESPACE, MAX_TIMEOUT
from ._pywbemcli_operations import PYWBEMCLIConnection, PYWBEMCLIFakedConnection

WBEM_SERVER_OBJ = None

PYWBEMCLI_LOG = 'pywbemcli.log'


def _raise_typeerror(name, value, rqd_type):
    """
    Generate a TypeError for a property in pywbem_server setter that has an
    invalid type
    """
    raise TypeError('Property "{0}" value: {1} must be type: "{2}", not type: '
                    '"{3}"'.format(name, value, rqd_type, type(value)))


def _validate_server_url(server):
    """
    Validate  and possibly complete the wbemserver url provided.

    Parameters:

      server (string):
        url of the WBEMServer to which connection is being made including
        scheme, hostname/IPAddress, and optional port

    Returns:
        The input url or url extended to use DEFAULT_SCHEME as the
        scheme if not is provided

    Raises:
        click.CLICKException if scheme invalid
    """
    if server[0] == '/':
        url = server

    elif re.match(r"^https{0,1}://", server) is not None:
        url = server

    elif re.match(r"^[a-zA-Z0-9]+://", server) is not None:
        raise click.ClickException('Invalid scheme on server argument: {}.'
                                   ' Use "http" or "https"'.format(server))
    else:
        url = "{scheme}://{host}".format(
            scheme=DEFAULT_URL_SCHEME,
            host=DEFAULT_URL_SCHEME)

    return url


class PywbemServer(object):
    # pylint: disable=too-many-instance-attributes, useless-object-inheritance
    """
    Envelope for connections with WBEM Server incorporates both the
    pywbem WBEMConnection and pywbem WBEMServer classes.

    The constructor is separated from the connection method to allow the
    object to be contstructed early but actually create the connection and
    WBEMServer objects only when required.  This allows parameters such
    as the password to be requested only when a connection is made.

    This class also holds the variables that determine whether the connection
    will use the pull operations or traditional operations
    """
    # ISSUE: #658 - Reorganize the location for these variable names separate
    # from the PywbemServer class.
    # The following class level variables are the names for the env variables
    # where server connection information are be saved and used as alternate
    # input sources for pywbemcli arguments and options.
    server_envvar = 'PYWBEMCLI_SERVER'
    name_envvar = 'PYWBEMCLI_NAME'
    user_envvar = 'PYWBEMCLI_USER'
    password_envvar = 'PYWBEMCLI_PASSWORD'
    defaultnamespace_envvar = 'PYWBEMCLI_DEFAULT_NAMESPACE'
    timeout_envvar = 'PYWBEMCLI_TIMEOUT'
    keyfile_envvar = 'PYWBEMCLI_KEYFILE'
    certfile_envvar = 'PYWBEMCLI_CERTFILE'
    verify_envvar = 'PYWBEMCLI_VERIFY'
    ca_certs_envvar = 'PYWBEMCLI_CA_CERTS'
    timestats_envvar = 'PYWBEMCLI_TIMESTATS'
    use_pull_envvar = 'PYWBEMCLI_USE_PULL'
    pull_max_cnt_envvar = 'PYWBEMCLI_PULL_MAX_CNT'
    mock_server_envvar = 'PYWBEMCLI_MOCK_SERVER'
    log_envvar = 'PYWBEMCLI_LOG'
    # The following exports are not part of the pywbem_server container
    pdb_envvar = 'PYWBEMCLI_PDB'
    connections_file_envvar = 'PYWBEMCLI_CONNECTIONS_FILE'

    def __init__(self, server=None, default_namespace=DEFAULT_NAMESPACE,
                 name='default', user=None, password=None,
                 timeout=DEFAULT_CONNECTION_TIMEOUT, verify=None, use_pull=None,
                 pull_max_cnt=None, certfile=None, keyfile=None,
                 ca_certs=None, mock_server=None, connections_file=None):
        """
        Create  a PywbemServer object. This contains the configuration
        and operation information to create a connection to the server
        and execute cim_operations on the server.

        Raises:
          TypeError: Type error in some input parameter.
          ValueError: Value error in some input parameter.
        """

        if server and mock_server:
            raise ValueError('Simultaneous "--server" and '
                             '"--mock-server" not allowed. Server: {}, '
                             'mock_server {}'.format(server, mock_server))
        self.server = server
        self.mock_server = mock_server
        self.name = name
        self.default_namespace = default_namespace
        self.user = user
        self.password = password
        self.timeout = timeout
        self.use_pull = use_pull
        self.pull_max_cnt = pull_max_cnt
        self.verify = verify
        self.certfile = certfile
        self.keyfile = keyfile
        self.ca_certs = ca_certs

        # May be None in case of not-saved connection (e.g. connection save)
        self._connections_file = connections_file

        # WBEMServer object when connected; None when disconnected.
        self._wbem_server = None

    def __str__(self):
        return 'PywbemServer(url={s._server} name={s.name})'.format(s=self)

    def __repr__(self):
        return 'PywbemServer(server={s._server} name={s.name} ' \
               'ns={s.default_namespace} user={s.user} ' \
               'password={s.password} timeout={s.timeout} ' \
               'use_pull={s.use_pull} pull_max_cnt={s.pull_max_cnt} ' \
               'verify={s.verify} certfile={s.certfile} keyfile={s.keyfile} ' \
               'ca_certs={s.ca_certs} mock_server={s.mock_server!r} ' \
               'wbem_server={s._wbem_server!r})'.format(s=self)

    @property
    def server(self):
        """
        :term:`string`: Scheme with Hostname or IP address of the WBEM Server.
        """
        return self._server

    @server.setter
    def server(self, server):
        """Setter method; for a description see the getter method."""

        # pylint: disable=attribute-defined-outside-init
        if server:
            self._server = _validate_server_url(server)
        else:
            self._server = server

    @property
    def mock_server(self):
        """
        list of :term:`string`: list of file paths for mock server setup.
        """
        return self._mock_server

    @mock_server.setter
    def mock_server(self, mock_server):
        """Setter method; for a description see the getter method."""
        if mock_server:
            if not isinstance(mock_server, (list, six.string_types)):
                _raise_typeerror("mock_server", mock_server, 'list, string')
            for ms in mock_server:
                if not isinstance(ms, six.string_types):
                    _raise_typeerror("mock_server item", mock_server, 'string')
        # assure this is list type in yaml output
        if mock_server is None:
            mock_server = []
        # pylint: disable=attribute-defined-outside-init
        self._mock_server = mock_server

    @property
    def name(self):
        """
        :term:`string`: Defines a name for this connection object.
        """
        return self._name

    @name.setter
    def name(self, name):
        """Setter method; for a description see the getter method."""
        if name:
            if not isinstance(name, six.string_types):
                _raise_typeerror("name", name, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._name = name

    @property
    def user(self):
        """
        :term:`string`: user name on the WBEM Server.
        """
        return self._user

    @user.setter
    def user(self, user):
        """Setter method; for a description see the getter method."""

        if user and not isinstance(user, six.string_types):
            _raise_typeerror("user", user, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._user = user

    @property
    def password(self):
        """
        :term:`string`: Password for this user on this WBEM Server.
        """
        return self._password

    @password.setter
    def password(self, password):
        """Setter method; for a description see the getter method."""
        if password and not isinstance(password, six.string_types):
            _raise_typeerror("password", password, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._password = password

    @property
    def default_namespace(self):
        """
        :term:`string`: Namespace to be used as default  for requests.
        """
        return self._default_namespace

    @default_namespace.setter
    def default_namespace(self, default_namespace):
        """Setter method; for a description see the getter method."""
        if default_namespace and not isinstance(default_namespace,
                                                six.string_types):
            _raise_typeerror("default-namespace", default_namespace, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._default_namespace = default_namespace

    @property
    def timeout(self):
        """
        :term:`int`: Connection timeout to be used on requests in seconds
        """
        if self._timeout:
            assert isinstance(self._timeout, int)
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Setter method; for a description see the getter method."""
        if timeout is None:   # disallow None
            raise ValueError('timeout option of None not allowed')
        if not isinstance(timeout, int):
            _raise_typeerror("timeout", timeout, 'integer')
        if not 0 < timeout <= MAX_TIMEOUT:
            raise ValueError('Timeout option "{0}" out of range {1} to {2} sec'
                             .format(timeout, 0, MAX_TIMEOUT))
        # pylint: disable=attribute-defined-outside-init
        self._timeout = timeout

    @property
    def use_pull(self):
        """
        :term:`string`: Choice of strings that defines action. The
        choices are "either", "yes", "no"
        """
        return self._use_pull

    @use_pull.setter
    def use_pull(self, use_pull):
        """Setter method; for a description see the getter method."""

        if use_pull is None or isinstance(use_pull, bool):
            # pylint: disable=attribute-defined-outside-init
            self._use_pull = use_pull
        else:
            _raise_typeerror("use-pull", use_pull, 'boolean')

    @property
    def pull_max_cnt(self):
        """
        :term:`int`: Count of number of objects to be retrieved with each
        pull operation or None if the default value is to be used
        """
        return self._pull_max_cnt

    @pull_max_cnt.setter
    def pull_max_cnt(self, pull_max_cnt):
        """Setter method; for a description see the getter method."""

        if pull_max_cnt is None or isinstance(pull_max_cnt, six.integer_types):
            # pylint: disable=attribute-defined-outside-init
            self._pull_max_cnt = pull_max_cnt
        else:
            _raise_typeerror("pull_max_cnt", pull_max_cnt, 'integer')

    @property
    def verify(self):
        """
        :class:`py:bool`: Connection server verfication flag. If True
        server cert verified during connection.
        """
        return self._verify

    @verify.setter
    def verify(self, verify):
        """Setter method; for a description see the getter method."""
        if verify and not isinstance(verify, bool):
            _raise_typeerror("verify", verify, 'boolean')
        # pylint: disable=attribute-defined-outside-init
        self._verify = verify

    @property
    def certfile(self):
        """
        :term:`string`: certificate for server or None if parameter not
        provided on input
        """
        if self._certfile:
            assert isinstance(self._certfile, six.string_types)
        return self._certfile

    @certfile.setter
    def certfile(self, certfile):
        """Setter method; for a description see the getter method."""
        if certfile and not isinstance(certfile, six.string_types):
            _raise_typeerror("certfile", certfile, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._certfile = certfile

    @property
    def keyfile(self):
        """
        :term:`string`: keyfile or None if no keyfile parameter input
        """
        if self._keyfile:
            assert isinstance(self._keyfile, six.string_types)
        return self._keyfile

    @keyfile.setter
    def keyfile(self, keyfile):
        """Setter method; for a description see the getter method."""
        if keyfile and not isinstance(keyfile, six.string_types):
            _raise_typeerror("keyfile", keyfile, 'string')

        # pylint: disable=attribute-defined-outside-init
        self._keyfile = keyfile

    @property
    def ca_certs(self):
        """
        :term:`string`: String that defines certs for server validation"
        """
        if self._ca_certs:
            assert isinstance(self._ca_certs, six.string_types)
        return self._ca_certs

    @ca_certs.setter
    def ca_certs(self, ca_certs):
        """Setter method; for a description see the getter method."""
        if ca_certs and not isinstance(ca_certs, six.string_types):
            _raise_typeerror("ca_certs", ca_certs, 'string')
        # pylint: disable=attribute-defined-outside-init
        self._ca_certs = ca_certs

    @property
    def conn(self):
        """
        :class:`PYWBEMCLIFakedConnection` or :class:`PYWBEMCLIConnection`
        (both derived from :class:`pywbem.WBEMConnection`):
        Connection object to be used for WBEM operation requests, when
        connected to the server.

        `None` when disconnected from the server.
        """
        return self.wbem_server.conn

    @property
    def wbem_server(self):
        """
        :class:`~pywbem.WBEMServer`: WBEMServer instance to be used for
        higher level WBEM server requests, when connected to the server.

        `None` when disconnected from the server.

        Using this property will connect to the server if not yet connected,
        and will raise an exception if no server is specified.

        When connecting to a real WBEM server, a password is prompted for if
        a user is specified in the current context but no password. The
        password is saved in the context, so the password is prompted for only
        once (e.g. in interactive mode).
        """
        ctx = click.get_current_context()
        if self._wbem_server is None:
            self.get_password(ctx.obj)
            self.connect(
                log=ctx.obj.log,
                use_pull=ctx.obj.use_pull,
                verbose=ctx.obj.verbose)
        return self._wbem_server

    @property
    def connected(self):
        """
        Current connected state. Returns True if connected and
        False if not connected.
        """
        return bool(self._wbem_server)

    def password_prompt(self, ctx):
        """
        Request password from console.
        """
        if self.user:
            ctx.spinner_stop()
            password = click.prompt(
                "Enter password (user {user})" .format(user=self.user),
                hide_input=True,
                confirmation_prompt=False, type=str, err=True)
            ctx.spinner_start()
            # pylint: disable=attribute-defined-outside-init
            self.password = password
        else:
            raise click.ClickException("{cmd} requires user/password, but "
                                       "no password provided."
                                       .format(cmd=ctx.invoked_subcommand))

    def get_password(self, ctx):
        """
        Conditional password prompt function. Prompts for password only if
        there is a defined user and no password.
        """
        if self.user and not self.password:
            self.password_prompt(ctx)

    def to_dict(self):
        """
        Create dictionary from instance for persisting the connection. All
        key names that include _ are modified to - so that the dictionary
        reflects the general option names.
        """

        return OrderedDict({"name": self.name,
                            "server": self.server,
                            "user": self.user,
                            "password": self.password,
                            "default-namespace": self.default_namespace,
                            "timeout": self.timeout,
                            "use_pull": self.use_pull,
                            "pull_max_cnt": self.pull_max_cnt,
                            "verify": self.verify,
                            "certfile": self.certfile,
                            "keyfile": self.keyfile,
                            "ca-certs": self.ca_certs,
                            "mock-server": self.mock_server})

    @staticmethod
    def create(replace_underscores=False, **kwargs):
        """
        Create PywbemServer object from kwargs. If replace_underscore is
        True, replace any -  in names with _

        Raises:
          TypeError: Type error in some input parameter.
          ValueError: Value error in some input parameter.
        """
        kwargsout = {}
        if replace_underscores:
            kwargsout = {k.replace('-', '_'): v for k, v in kwargs.items()}
        # Test for existence of required elements
        kwargsout['name']   # pylint: disable=pointless-statement

        # Normalize file names for the server environment
        if 'mock_server' in kwargsout:
            if kwargsout['mock_server']:
                kwargsout['mock_server'] = [os.path.normpath(fn) for fn in
                                            kwargsout['mock_server']]
        return PywbemServer(**kwargsout)

    def disconnect(self):
        """
        Disconnect from the server, closing the connection.

        Must be connected to the server when calling this method.
        """
        assert self._wbem_server is not None  # Must be connected

        ctx = click.get_current_context()
        if ctx.obj:
            if ctx.obj.verbose:
                if self._mock_server:
                    server_txt = "mock environment {}".format(self._mock_server)
                else:
                    server_txt = "WBEM server {}".format(self._server)
                click.echo("Disconnecting from {}".format(server_txt))

        self._wbem_server.conn.close()
        self._wbem_server = None

    def connect(self, log=None, use_pull=None, verbose=None):
        """
        Connect to the server, using the current attributes of this object.

        Must be disconnected from the server when calling this method.

        If `self.mock_server` is non-empty, a mock connection is created.
        Otherwise, `self.server` must be set and a real connection is created.

        Raises:
          ClickException: Several issues that cause the command (the whole
            command in command mode, or a single command in interactive mode)
            to terminate.
          ValueError: Other issues with server-related attributes of this
            object that are considered programming errors.
        """

        assert self._wbem_server is None  # Must be disconnected

        ctx = click.get_current_context()
        if ctx.obj:
            if ctx.obj.verbose:
                if self._mock_server:
                    server_txt = "mock environment {}".format(self._mock_server)
                else:
                    server_txt = "WBEM server {}".format(self._server)
                click.echo("Connecting to {}".format(server_txt))

        if self._mock_server:
            conn = PYWBEMCLIFakedConnection(
                default_namespace=self.default_namespace,
                use_pull_operations=use_pull,
                stats_enabled=True)
            self._wbem_server = WBEMServer(conn)
            try:
                conn.build_mockenv(
                    self._wbem_server, self._mock_server,
                    self._connections_file, self._name, verbose)
            except (mockscripts.MockMOFCompileError,
                    mockscripts.MockScriptError) as exc:
                # These errors cause the command to be aborted, because
                # partially executed mock scripts or partially compiled MOF
                # files might have caused inconsistencies in the Python
                # mockscripts namespace and in the CIM repository.
                click.echo(str(exc), err=True)
                raise click.Abort()
            except mockscripts.MockError as exc:
                raise click.ClickException(str(exc))
        else:  # mock_server does not exist
            if not self.server:
                raise click.ClickException('No server found. Cannot '
                                           'connect.')
            if self.keyfile is not None and self.certfile is None:
                ValueError('keyfile option requires certfile option')

            creds = (self.user, self.password) if self.user else None

            # If client cert and key provided, create dictionary for
            # wbem connection certs (WBEMConnection takes dictionary for this
            # info)
            x509_dict = None
            if self.certfile is not None:
                x509_dict = {"cert_file": self.certfile}
                if self.keyfile is not None:
                    x509_dict.update({'key_file': self.keyfile})

            # Create the WBEMConnection object and the _wbem_server object

            # Negate verify to no_verification
            if self.verify is None:
                no_verification = self.verify
            else:
                no_verification = not self.verify

            # Convert ca_certs command line option to ca_certs parameter
            if getattr(pywbem, 'PYWBEM_USES_REQUESTS', False):
                if self.ca_certs == 'certifi':
                    ca_certs = None
                else:
                    ca_certs = self.ca_certs
            else:
                ca_certs = self.ca_certs

            try:
                conn = PYWBEMCLIConnection(
                    self.server, creds,
                    default_namespace=self.default_namespace,
                    no_verification=no_verification,
                    x509=x509_dict, ca_certs=ca_certs,
                    timeout=self.timeout,
                    use_pull_operations=use_pull,
                    stats_enabled=True)
            except IOError as exc:
                raise click.ClickException(
                    'Cannot create connection to {}: {}'.
                    format(self.server, exc))

            # Create a WBEMServer object
            self._wbem_server = WBEMServer(conn)

        if log:
            try:
                configure_loggers_from_string(log,
                                              log_filename=PYWBEMCLI_LOG,
                                              connection=conn, propagate=True)
            except ValueError as ve:
                raise click.ClickException('Logger configuration error. input: '
                                           '{}. Exception: {}'.format(log, ve))



"""
This file contains extensions to Click specifically for pywbemcli
"""
import sys
from collections import OrderedDict

import click

from ._common import GENERAL_OPTS_TXT, CMD_OPTS_TXT


class PywbemcliGroup(click.Group):
    """
    Extend Click Group class to:
    1.  Order the display of commands within the help.
        The commands are ordered in the order that their definitions
        appears in the source code for each command group.
    This extension has a general name because it may be used for more than
    one extension to the Click.Group class.

    2. Force use of our method pywbemcli_format_usage rather than
       the click define Command.format_usage()

    It is used on all group decorators (@click.group(...)) except for the
    top level in pywbemcli which uses the PywbemcliTopGroup class.
    """

    # Use ordered dictionary to sort commands by their order defined in the
    # _cmd_... source file.
    def __init__(self, name=None, commands=None, **attrs):
        """
        Use OrderedDict to keep order commands inserted into command dict.
        Only required for Python versions that do not order dictionaries.
        Must be set after calling superclass __inits_ because click forces
        {} for this variable even if user were to set commands to OrderedDict
        """
        super(PywbemcliGroup, self).__init__(name, commands, **attrs)

        if sys.version_info < (3, 6):
            self.commands = commands or OrderedDict()

        # Replace Click.Command.format_usage with local version
        click.core.Command.format_usage = pywbemcli_format_usage
        click.core.Command.format_options = pywbemcli_format_options

    def list_commands(self, ctx):
        """
        Replace click.list_commands to eliminate the sort of cmd names.
        """
        return self.commands.keys()


class PywbemcliTopGroup(click.Group):
    """
    Extensions to be used with the top level help (pywbemcli --help)

    Extend Click Group class to:
    1.  Order the display of the commands and command groups in the top level
        help output to sort and then put names defined in the predefined_list
        at the end of the list of commands/groups. Since ordering of the top
        level cannot be tied to order commands are inserted in list, we elected
        to just move the generic ones to the end of the list.

    This extension has a general name because it may be used for more than
    one extension to the Click.Group class.
    """

    def __init__(self, name=None, commands=None, **attrs):
        """
        Use OrderedDict to keep order commands inserted into command dict.
        Only required for Python versions that do not order dictionaries.
        Must be set after calling superclass __inits_ because click forces
        {} for this variable even if user were to set commands to OrderedDict
        """
        super(PywbemcliTopGroup, self).__init__(name, commands, **attrs)

        # Replace Click.Command.format_options with local version
        click.core.Command.format_options = pywbemcli_format_options

    def list_commands(self, ctx):
        """
        Order The top level commands by sorting and then moving any commands
        defined in  move_to_end list to the end of the list.  This is
        the class override for click.Group ONLY with the top group, the
        click startup group in pywbemcli because in reorders the commands
        display specifically as we want them for pywbemcli.
        """
        # tuple of commands to move to bottom after sort
        move_to_end = ('connection', 'help', 'repl')

        # Sort because thier is no particular order for the groups
        cmd_list = sorted(self.commands.keys())
        pop_count = 0
        # reorder list so the move_to_end list commands are at bottom
        for i in range(len(cmd_list)):
            if cmd_list[i - pop_count] in move_to_end:
                cmd_list.append(cmd_list.pop(i - pop_count))
                pop_count += 1
        return cmd_list


class PywbemcliCommand(click.Command):
    """
    Modify the command usage formatter to show the commands in a format
    that fits how we use the tool. This should be the cls= argument
    for all pywbemcli command decorators (i.e. `@class_group.command(`).
    """

    def format_usage(self, ctx, formatter):
        """
        Replaces click.Command.format_usage
        """
        pieces = self.collect_usage_pieces(ctx)
        cmd_paths = ctx.command_path.split()
        # Reorder pieces to order GENERAL_OPTS_TXT, <cmd> ...
        new_pieces = [GENERAL_OPTS_TXT] + cmd_paths[1:] + pieces[1:] \
            + [pieces[0]]
        formatter.write_usage(cmd_paths[0], " ".join(new_pieces))


def pywbemcli_format_usage(self, ctx, formatter):
    """
    Replaces click.Command.format_usage in the fixes for click.Group
    """
    pieces = self.collect_usage_pieces(ctx)
    cmd_paths = ctx.command_path.split()

    # cmd_path is [<appname>, group, cmd]
    # If cmd_path has multiple components we are showing
    # app name, CMDGRP <cmd>. Break out CMDGRP and <cmd>
    # and reinsert after pieces[0] (the app name)
    if len(cmd_paths) > 1:
        new_pieces = [pieces[0]] + cmd_paths[1:] + pieces[1:]
        formatter.write_usage(cmd_paths[0], " ".join(new_pieces))
    else:
        formatter.write_usage(ctx.command_path, " ".join(pieces))


def pywbemcli_format_options(self, ctx, formatter):
    """
    Writes all the options into the formatter if they exist.
    """
    opts = []
    cmds_len = len(ctx.command_path.split())

    # If group or command level, use CMD_OPTS_TXT
    options_title = CMD_OPTS_TXT if cmds_len > 1 else GENERAL_OPTS_TXT
    # Cvt to string of words with first char of each word capitalized
    options_title = options_title[1:-1].lower().replace('-', " ").title()

    for param in self.get_params(ctx):
        rv = param.get_help_record(ctx)
        if rv is not None:
            opts.append(rv)

    if opts:
        with formatter.section(options_title):
            formatter.write_dl(opts)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Mixin class that adds methods to WBEMConnection and FakeWBEMConnection for
pywbemcli usage

This contains only methods that use the iter<...> operations  but also execute
the complete iterations so that we can use these as common operations for
pywbemcli instead of having to execute an algorithm of pull vs non-pull
everywhere xa WBEMConnection possible pull operation is called.

It also adds a method to FakeWBEMConnection to build the repository.
"""

from __future__ import absolute_import, print_function

import os
import errno
import glob
import hashlib
import pickle
import click
import packaging.version
import pywbem
import pywbem_mock

from .config import DEFAULT_MAXPULLCNT
from ._utils import ensure_bytes, DEFAULT_CONNECTIONS_FILE
from . import mockscripts

PYWBEM_VERSION = packaging.version.parse(pywbem.__version__)


#  __all__ = ['PYWBEMCLIConnection', 'PYWBEMCLIFakedConnection']


# pylint: disable=useless-object-inheritance
class PYWBEMCLIConnectionMixin(object):
    """
    Mixin class to extend WBEMConnection with a set of methods that use the
    iter<...> methods as the basis for getting Instances, etc. but add the
    generator processing to retrieve the instances.  These can be used within
    pywbemcli to allow one method call to ack as either a pull or traditional
    operation pushing the differences into this mixin.

    These methods do not resolve the core issues between the traditional and
    pull operations such as the fact that only the pull operations pass
    the FilterQuery parameter.

    They are a pywbemcli convience to simplify the individual action processing
    methods to a single call.
    """
    def PyWbemcliEnumerateInstancePaths(self, ClassName, namespace=None,
                                        FilterQueryLanguage=None,
                                        FilterQuery=None,
                                        OperationTimeout=None,
                                        ContinueOnError=None,
                                        MaxObjectCount=DEFAULT_MAXPULLCNT,
                                        **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterEnumerateInstancePaths and retrieve the instances. Returns
        the returned instances.

        Uses the same parameters as the IterEnumerateInstancePaths method.

        All exceptions from the underlying command are passed through this
        method.
        """

        result = self.IterEnumerateInstancePaths(
            ClassName,
            namespace=namespace,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliEnumerateInstances(self, ClassName, namespace=None,
                                    LocalOnly=None,
                                    DeepInheritance=None,
                                    IncludeQualifiers=None,
                                    IncludeClassOrigin=None, PropertyList=None,
                                    FilterQueryLanguage=None, FilterQuery=None,
                                    OperationTimeout=None, ContinueOnError=None,
                                    MaxObjectCount=DEFAULT_MAXPULLCNT,
                                    **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterEnumerateInstances and retrieve the instances. Returns
        the returned instances.

        Uses the same parameters as the IterEnumerateInstances method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterEnumerateInstances(
            ClassName,
            namespace=namespace,
            LocalOnly=LocalOnly,
            DeepInheritance=DeepInheritance,
            IncludeQualifiers=IncludeQualifiers,
            IncludeClassOrigin=IncludeClassOrigin,
            PropertyList=PropertyList,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliReferenceInstancePaths(self, InstanceName, ResultClass=None,
                                        Role=None,
                                        FilterQueryLanguage=None,
                                        FilterQuery=None,
                                        OperationTimeout=None,
                                        ContinueOnError=None,
                                        MaxObjectCount=DEFAULT_MAXPULLCNT,
                                        **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterReferemceInstancePaths and retrieve the instances. Returns
        the paths that result from iterating the IterReferenceInstancePaths.

        Uses the same parameters as the IterReferemceInstancePaths method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterReferenceInstancePaths(
            InstanceName,
            ResultClass=ResultClass,
            Role=Role,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliReferenceInstances(self, InstanceName, ResultClass=None,
                                    Role=None, IncludeQualifiers=None,
                                    IncludeClassOrigin=None, PropertyList=None,
                                    FilterQueryLanguage=None, FilterQuery=None,
                                    OperationTimeout=None, ContinueOnError=None,
                                    MaxObjectCount=DEFAULT_MAXPULLCNT,
                                    **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterReferencesInstances and retrieve the instances. Returns
        the returned instances.

        Uses the same parameters as the IterReferencesInstances method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterReferenceInstances(
            InstanceName,
            ResultClass=ResultClass,
            Role=Role,
            IncludeQualifiers=IncludeQualifiers,
            IncludeClassOrigin=IncludeClassOrigin,
            PropertyList=PropertyList,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliAssociatorInstancePaths(self, InstanceName, AssocClass=None,
                                         ResultClass=None,
                                         Role=None, ResultRole=None,
                                         FilterQueryLanguage=None,
                                         FilterQuery=None,
                                         OperationTimeout=None,
                                         ContinueOnError=None,
                                         MaxObjectCount=DEFAULT_MAXPULLCNT,
                                         **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterAssociatorInstancePaths and retrieve the paths. Returns
        the paths that result from iterating the IterAssociatorInstancePaths.

        Uses the same parameters as the IterAssociatorInstancePaths method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterAssociatorInstancePaths(
            InstanceName,
            AssocClass=AssocClass,
            ResultClass=ResultClass,
            Role=Role,
            ResultRole=ResultRole,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliAssociatorInstances(self, InstanceName, AssocClass=None,
                                     ResultClass=None,
                                     Role=None, ResultRole=None,
                                     IncludeQualifiers=None,
                                     IncludeClassOrigin=None, PropertyList=None,
                                     FilterQueryLanguage=None, FilterQuery=None,
                                     OperationTimeout=None,
                                     ContinueOnError=None,
                                     MaxObjectCount=DEFAULT_MAXPULLCNT,
                                     **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterAssociatorInstances and retrieve the instances. Returns
        the instances that result from iterating the IterAssociatorInstances.

        Uses the same parameters as the IterAssociatorInstances method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterAssociatorInstances(
            InstanceName,
            AssocClass=AssocClass,
            ResultClass=ResultClass,
            Role=Role,
            ResultRole=ResultRole,
            IncludeQualifiers=IncludeQualifiers,
            IncludeClassOrigin=IncludeClassOrigin,
            PropertyList=PropertyList,
            FilterQueryLanguage=FilterQueryLanguage,
            FilterQuery=FilterQuery,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)

    def PyWbemcliQueryInstances(self, FilterQueryLanguage, FilterQuery,
                                namespace=None, ReturnQueryResultClass=None,
                                OperationTimeout=None, ContinueOnError=None,
                                MaxObjectCount=DEFAULT_MAXPULLCNT,
                                **extra):
        # pylint: disable=unused-argument
        # pylint: disable=invalid-name
        """
        Execute IterQueryInstances and retrieve the instances. Returns
        the instances that result from iterating the IterQueryInstances.

        Uses the same parameters as the IterQueryInstances method.

        All exceptions from the underlying method are passed through this
        method.
        """

        result = self.IterQueryInstances(
            FilterQueryLanguage,
            FilterQuery,
            namespace=namespace,
            ReturnQueryResultClass=ReturnQueryResultClass,
            OperationTimeout=OperationTimeout,
            ContinueOnError=ContinueOnError,
            MaxObjectCount=MaxObjectCount)
        return list(result)


class BuildMockenvMixin(object):
    # pylint: disable=too-few-public-methods
    """
    Mixin class for pywbem_mock.FakedWBEMConnection that adds the ability to
    build the mock environment of a connection from a connection definition in
    a connections file.
    """

    def build_mockenv(self, server, file_path_list, connections_file,
                      connection_name, verbose):
        """
        Builds the mock environment of the 'self' connection from the input
        files, or from the mock cache of the connection if it is up to date.
        If the mock environment was built from the input files, the mock
        environment of the connection is dumped to its cache.

        The input files for building the mock environment are:

        * MOF files with a suffix of '.mof'.

          These files are compiled into the default namespace of the connection.

        * Python files with a suffix of '.py'.

          These files are mock scripts that are imported and thereby executed.
          The mock scripts can be used for any kind of setup of the mock
          environment, for example for creating namespaces, for defining
          provider classes and registering providers, or for adding CIM objects
          either directly through add_cimobjects() or by compiling MOF files.

          Mock scripts support two approaches for passing the connection and
          server objects they should operate on:

          * via a setup() function defined in the mock script. This is the
            recommended approach, and it supports caching. The setup()
            function has the following parameters:

              conn (pywbem_mock.FakedWBEMConnection): The mock connection.

              server (pywbem.WBEMServer): The server object for the mock
                connection.

              verbose (bool): Verbose flag from the command line.

          * via global variables made available to the mock script. This
            approach prevents caching. The following global variables are
            made available:

              CONN (pywbem_mock.FakedWBEMConnection): The mock connection.

              SERVER (pywbem.WBEMServer): The server object for the mock
                connection.

              VERBOSE (bool): Verbose flag from the command line.

        Parameters:

          self (pywbem_mock.FakedWBEMConnection): The mock connection.

          server (pywbem.WBEMServer): The server object for the mock connection.

          file_path_list (list of string): The path names of the input files
            for building the mock environment, from the connection definition.

          connections_file (string): Path name of the connections file.

          connection_name (string): The name of the connection definition in
            the connections file.

          verbose (bool): Verbose flag from the command line.

        Raises:
          MockFileError: Mock file does not exist.
          MockMOFCompileError: Mock MOF file fails to compile.
          MockScriptError: Mock script fails to execute.
          SetupNotSupportedError (py<3.5): New-style setup in mock script not
            supported.
        """

        # Check that the input files exist. Since we loop through them multiple
        # times, we check that once.
        for file_path in file_path_list:
            if not os.path.exists(file_path):
                raise mockscripts.MockFileError(
                    "Mock file does not exist: {}".format(file_path))

        # The connections file is set if a named connection is used, i.e.
        # when specifying the -n general option. It is not set when the -s or -m
        # general options were specified. When no connections file is set, no
        # caching happens because there is no connection definition context
        # which is required for caching.
        if connections_file == DEFAULT_CONNECTIONS_FILE:

            cache_rootdir = mockcache_rootdir()
            if not os.path.isdir(cache_rootdir):
                os.mkdir(cache_rootdir)

            cache_dir = mockcache_cachedir(
                cache_rootdir, connections_file, connection_name)
            if not os.path.isdir(cache_dir):
                os.mkdir(cache_dir)

            # The mockenv pickle file contains the pickled state of the mock
            # environment.
            mockenv_pickle_file = os.path.join(cache_dir, 'mockenv.pkl')

            # The depreg pickle file contains the provider dependents
            # registry of the connection. It is used to look up the dependent
            # files of a mock script. The content of these dependent files is
            # also taken into account when determining whether the cache is up
            # to date. This needs to go into a separate pickle file because
            # it needs to be loaded and examined before the mckenv pickle
            # file is loaded.
            depreg_pickle_file = os.path.join(cache_dir, 'depreg.pkl')

            # The md5 file contains the MD5 hash value of the content of the
            # input files for the mock environment, and also taken into account
            # when determining whether the cache is up to date.
            md5_file = os.path.join(cache_dir, 'mockfiles.md5')

            # Flag indicating that the mock environment needs to be built
            # (or re-built). If False, the mock environment cache can be used.
            need_rebuild = False

            # Determine whether the mock environment needs to be rebuilt based
            # on the (non-)existence of the cache files.
            if not os.path.isfile(mockenv_pickle_file) \
                    or not os.path.isfile(depreg_pickle_file) \
                    or not os.path.isfile(md5_file):
                if verbose:
                    click.echo("Mock environment for connection definition "
                               "'{}' will be built because it was not cached.".
                               format(connection_name))
                need_rebuild = True

            try:
                depreg = self._load_depreg(depreg_pickle_file)
            except (IOError, OSError) as exc:
                if exc.errno == errno.ENOENT:
                    depreg = pywbem_mock.ProviderDependentRegistry()
                else:
                    raise

            # Calculate the MD5 hash value of the content of the input files
            md5 = hashlib.md5()
            for file_path in file_path_list:
                with open(file_path, 'rb') as fp:
                    file_source = fp.read()
                    md5.update(file_source)

                # For mock scripts, take their dependent files into account
                if file_path.endswith('.py'):
                    dep_files = depreg.iter_dependents(file_path)
                    for dep_file in dep_files:
                        with open(dep_file, 'rb') as fp:
                            file_source = fp.read()
                            md5.update(file_source)

            # Add the cache dir, so that manual tweaks on the cache files
            # invalidates the cache.
            md5.update(ensure_bytes(cache_dir))

            new_md5_value = md5.hexdigest()

            # Determine whether the mock environment needs to be rebuilt based
            # on the MD5 hash value of the input file content.
            if not need_rebuild:
                with open(md5_file, 'r') as fp:
                    cached_md5_value = fp.read()
                if new_md5_value != cached_md5_value:
                    if verbose:
                        click.echo("Mock environment for connection "
                                   "definition '{}' is cached but will be "
                                   "rebuilt because the mock files have "
                                   "changed.".format(connection_name))
                    need_rebuild = True

            cache_it = True

        elif connections_file:
            # User-specified connections file used.

            if verbose:
                click.echo("Mock environment for connection definition '{}' "
                           "will be built because user-specified connections "
                           "files are not cached.".format(connection_name))
            need_rebuild = True
            cache_it = False

        else:
            # No connections file context.

            if verbose:
                click.echo("Mock environment for connection definition '{}' "
                           "will be built because no connections file is "
                           "known.".format(connection_name))
            need_rebuild = True
            cache_it = False

        if need_rebuild:
            try:
                self._build_mockenv(server, file_path_list, verbose)
            except mockscripts.NotCacheable as exc:
                if verbose:
                    click.echo("Mock environment for connection definition "
                               "'{}' will be built because it is not "
                               "cacheable: {}.".format(connection_name, exc))
            else:
                if connections_file and cache_it:
                    self._dump_mockenv(mockenv_pickle_file)
                    self._dump_depreg(
                        self.provider_dependent_registry, depreg_pickle_file)
                    with open(md5_file, 'w') as fp:
                        fp.write(new_md5_value)
                    if verbose:
                        click.echo("Mock environment for connection "
                                   "definition '{}' has been written to "
                                   "cache.".format(connection_name))
        else:
            # When no rebuild is needed, there must have been a connections
            # file set.
            assert connections_file
            try:
                self._load_mockenv(mockenv_pickle_file, file_path_list)
                if verbose:
                    click.echo("Mock environment for connection definition "
                               "'{}' has been loaded from cache.".
                               format(connection_name))
            except mockscripts.NotCacheable as exc:
                if verbose:
                    click.echo("Mock environment for connection definition "
                               "'{}' will be rebuilt because it is not "
                               "cacheable: {}.".format(connection_name, exc))
                self._build_mockenv(server, file_path_list, verbose)

    def _build_mockenv(self, server, file_path_list, verbose):
        """
        Build the mock environment from the input files.

        Parameters:

          self (pywbem_mock.FakedWBEMConnection): The mock connection.

          server (pywbem.WBEMServer): The server object for the mock connection.

          file_path_list (list of string): The path names of the input files
            for building the mock environment, from the connection definition.

          verbose (bool): Verbose flag from the command line.

        Raises:
          NotCacheable (py<3.5): Mock environment is not cacheable.
          MockMOFCompileError: Mock MOF file fails to compile.
          MockScriptError: Mock script fails to execute.
          SetupNotSupportedError (py<3.5): New-style setup in mock script not
            supported.
        """
        for file_path in file_path_list:
            ext = os.path.splitext(file_path)[1]
            if ext == '.mof':
                try:
                    # Displays any MOFParseError already
                    self.compile_mof_file(file_path, verbose=verbose)
                except pywbem.Error as er:
                    # Abort the entire pywbemcli command because the
                    # MOF compilation might have caused inconsistencies in
                    # the mock repository.

                    if PYWBEM_VERSION.release >= (1, 0, 0):
                        # display just the exception.
                        msg = "MOF compile failed:\n{0}".format(er)
                    else:
                        # display file name.  Error text displayed already.
                        if isinstance(er, pywbem.MOFParseError):
                            msg = "MOF compile failed: File: '{0}'" \
                                "(see above)".format(file_path)
                        else:  # not parse error, display exception
                            msg = "MOF compile failed: File: {0} " \
                                "Error: {1}".format(file_path, er)
                    new_exc = mockscripts.MockMOFCompileError(msg)
                    new_exc.__cause__ = None
                    raise new_exc
            else:
                assert ext == '.py'  # already checked

                # May raise various mockscripts.MockError exceptions.
                # NotCacheable will be handled by the caller by building the
                # mock env.
                mockscripts.setup_script(file_path, self, server, verbose)

    def _dump_mockenv(self, mockenv_pickle_file):
        """
        Dump the mock environment of the connection to the mockenv pickle file.

        Parameters:

          self (pywbem_mock.FakedWBEMConnection): The mock connection.

          mockenv_pickle_file (pywbem.WBEMServer): Path name of the mockenv
            pickle file.
        """

        # Save the provider registry and the CIM repository

        # We construct a single object, because the CIM repository is
        # referenced from each provider, and pickle properly handles
        # multiple references to the same object.
        mockenv = dict(
            cimrepository=self.cimrepository,
            # pylint: disable=protected-access
            provider_registry=self._provider_registry,
        )
        with open(mockenv_pickle_file, 'wb') as fp:
            pickle.dump(mockenv, fp)

    def _load_mockenv(self, mockenv_pickle_file, file_path_list):
        """
        Load the mock environment from the mockenv pickle file.

        This method also imports the Python scripts from the input files in
        order to re-establish any class definitions that may be needed, for
        example provider classes.

        Parameters:

          self (pywbem_mock.FakedWBEMConnection): The mock connection.

          mockenv_pickle_file (pywbem.WBEMServer): Path name of the mockenv
            pickle file.

          file_path_list (list of string): The path names of the input files
            for building the mock environment, from the connection definition.

        Raises:
          NotCacheable (py<3.5): Mock environment is not cacheable.
        """

        # Restore the provider classes
        for file_path in file_path_list:
            ext = os.path.splitext(file_path)[1]
            if ext == '.py':
                # May raise mockscripts.NotCacheable which will be handled by
                # the caller by building the mock env.
                mockscripts.import_script(file_path)

        # Restore the provider registry and the CIM repository
        with open(mockenv_pickle_file, 'rb') as fp:
            mockenv = pickle.load(fp)

        # Others have references to the self._cimrepository object, so we are
        # not replacing that object, but are rather replacing the state of
        # that object.
        cimrepository = mockenv['cimrepository']
        assert isinstance(cimrepository, pywbem_mock.InMemoryRepository)
        # pylint: disable=protected-access
        self._cimrepository.load(cimrepository)

        provider_registry = mockenv['provider_registry']
        assert isinstance(provider_registry, pywbem_mock.ProviderRegistry)
        # pylint: disable=protected-access
        self._provider_registry.load(provider_registry)

    @staticmethod
    def _dump_depreg(depreg, depreg_pickle_file):
        """
        Dump a provider dependent registry to a pickle file.

        Parameters:

          depreg (pywbem_mock.ProviderDependentRegistry): Provider dependent
            registry to be dumped.

          depreg_pickle_file (string): Path name of the pickle file.
        """
        with open(depreg_pickle_file, 'wb') as fp:
            pickle.dump(depreg, fp)

    @staticmethod
    def _load_depreg(depreg_pickle_file):
        """
        Load a provider dependent registry from a pickle file and return it.

        Parameters:

          depreg_pickle_file (string): Path name of the pickle file to be
            loaded.

        Returns:
          pywbem_mock.ProviderDependentRegistry: Provider dependent registry.
        """
        with open(depreg_pickle_file, 'rb') as fp:
            depreg = pickle.load(fp)
        return depreg


class PYWBEMCLIConnection(pywbem.WBEMConnection, PYWBEMCLIConnectionMixin):
    """
    PyWBEMCLIConnection subclass adds the methods added by
    PYWBEMCLIConnectionMixin
    """

    def __init__(self, *args, **kwargs):
        """
        ctor passes all input parameters to superclass
        """
        super(PYWBEMCLIConnection, self).__init__(*args, **kwargs)


class PYWBEMCLIFakedConnection(BuildMockenvMixin,
                               PYWBEMCLIConnectionMixin,
                               pywbem_mock.FakedWBEMConnection):
    """
    PyWBEMCLIFakedConnection subclass adds the methods added by
    PYWBEMCLIConnectionMixin
    """
    def __init__(self, *args, **kwargs):
        """
        ctor passes all input parameters to superclass
        """
        super(PYWBEMCLIFakedConnection, self).__init__(*args, **kwargs)


def mockcache_rootdir():
    """
    Return the directory path of the mock cache root directory.
    """
    dir_path = os.path.join(os.path.expanduser('~'), '.pywbemcli_mockcache')
    return dir_path


def mockcache_cachedir(rootdir, connections_file, connection_name):
    """
    Return the directory path of the mock cache directory for a connection.
    """
    # Construct a (reproducible) cache ID from connections file path and
    # connection definition name.
    # Example: 6048a3da1a34a3ec605825a1493c7bb5.simple
    try:
        connections_file = os.path.relpath(
            connections_file, os.path.expanduser('~'))
    except ValueError:
        # On Windows, os.path.relpath() raises ValueError when the paths
        # are on different drives
        pass
    md5 = hashlib.md5()
    md5.update(connections_file.encode("utf-8"))
    cache_id = "{}.{}".format(md5.hexdigest(), connection_name)
    dir_path = os.path.join(rootdir, cache_id)
    return dir_path


def delete_mock_cache(connections_file, connection_name):
    """
    Delete the mock cache of the connection, if it exists.

    Parameters:

      self (pywbem_mock.FakedWBEMConnection): The mock connection.

      connections_file (string): Path name of the connections file.

      connection_name (string): The name of the connection definition in
        the connections file.

    Raises:
      OSError: Mock cache cannot be deleted.
    """
    cache_dir = mockcache_cachedir(
        mockcache_rootdir(), connections_file, connection_name)
    if os.path.isdir(cache_dir):
        file_list = glob.glob(os.path.join(cache_dir, '*'))
        for _file in file_list:
            os.remove(_file)
        os.rmdir(cache_dir)



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions to build a tree dictionary and also a displayable tree in ascii
from a list of CIM classes.  This code implements the ability to add the
information on the Association, Indication, Abstract, and Version qualifiers to
each classname (see show_detail parameter).
"""

from __future__ import absolute_import, print_function

import six
from asciitree import LeftAligned
import click

# Use an ordered Nocase dictionary for the tree. Ordered dictionary creates
# tree output that has the same order in multiple versions of python.
from pywbem._nocasedict import NocaseDict


def display_class_tree(classes, top_classname=None, show_detail=None):
    """
    Build and display in ASCII output a tree of classes. If show_detail
    is set, add class qualifier to each classname as part of
    the display

    Parameters:
        classes (list of :class:`~pywbem.CIMClass`)

        top_classname (:term: `string`)
            The top level class to display or None if the display is
            from root of the class hierarchy

        show_detal (:class": `boolean)
            Flag to indicate that more detail about each class be displayed

    """
    def build_details(klass):
        """
        Build a string of any of the qualifiers defined below.

        Returns:
          string defining the qualifiers defined.
        """
        def append_if(append_values):
            """Append the values if the append_values string exists """
            if append_values:
                rtn_values.append(append_values)

        def bool_qualifier_is(qname, qvalue):
            """
            If boolean qualifier qname exists and value matches qvalue
            return the name of the qualifier if the qualifier exists and
            matches qvalue or None if it does not exist or the value does
            not match qvalue
            """
            if qname in klass.qualifiers and \
                    klass.qualifiers[qname].value == qvalue:
                return qname

            return None

        def string_qualifier_value(qname):
            """
            Get the Version qualifier if it exists and return the value.
            If there is no Version qualifier return "".
            Returns the value or None.
            """
            q = klass.qualifiers.get(qname, None)
            if q is None:
                return None
            v = q.value
            return "{}={}".format(qname, v) if v else None

        # Build the rtn_values string from the following qualifiers and their
        # values
        rtn_values = []
        append_if(bool_qualifier_is('Association', True))
        append_if(bool_qualifier_is('Abstract', True))
        append_if(bool_qualifier_is('Indication', True))
        append_if(string_qualifier_value('Version'))
        return ', '.join(rtn_values)

    def extend_name(cln, classes_dict):
        """
        Fix the classname defined by cln by suffixing with the extended
        components for display of type and version.
        """
        if cln in classes_dict and show_detail:
            klass = classes_dict[cln]
            details = build_details(klass)
            new_cln = "{0} ({1})".format(cln, details)
            return new_cln

        return cln

    def fix_names(classname, classes_dict, flat_cln_subclns):
        """
        Modify the names in the key and list of values to the classname
        entity  that will be presented

        Parameters:

          classname (:term:`string`): Name of class to be processed

          classes_dict (dict): CIMClass objects by class name
        """
        new_key = extend_name(classname, classes_dict)
        new_values = []
        for cln in flat_cln_subclns[classname]:
            new_values.append(extend_name(cln, classes_dict))
        return new_key, new_values

    # Start of display_class_tree method

    cln_to_supercln = build_cln_to_supercln_dict(classes)

    # If top_class is none, create artifical root as superclassname from
    # entries with value None.
    top_classname_default = 'root'
    if top_classname is None:
        for cln in cln_to_supercln:
            if not cln_to_supercln[cln]:
                cln_to_supercln[cln] = top_classname_default

    flat_cln_subclns = build_subcln_in_cln(cln_to_supercln)

    # If the names are to be expanded with additional information, rebuild
    # flat_cln_subclns_dict and change all the names.

    fixed_dict = NocaseDict()
    if show_detail:
        classes_dict = {cls.classname: cls for cls in classes}
        for cln in flat_cln_subclns:
            new_name, new_values = fix_names(cln, classes_dict,
                                             flat_cln_subclns)
            fixed_dict[new_name] = new_values
        flat_cln_subclns = fixed_dict

    tcv = top_classname if top_classname else top_classname_default
    # build nested tree from the flat list of subclasses_lists in classes
    nested_class_tree = build_nested_tree(flat_cln_subclns, tcv)

    tr = LeftAligned()
    click.echo(tr(nested_class_tree))


def build_cln_to_supercln_dict(classes):
    """
    Build a dictionary of classname: superclassname and sort it by
    to order the keys from a list of CIM classes.

      Parameters:

        classes (list of :class:`~pywbem.CIMClass`)
            List of classes to be included in the tree including at least
            the classname and superclass name.  All other parameters are
            ignored

      Returns:
        NocaseDict of classes where the keys are all the classes and
        the values are either the superclass name for the class or None if
        there is no superclass
    """

    # Build dictionary of classname : superclassname from list of CIM classes
    cln_to_supercln = {cln.classname: cln.superclass for cln in classes}

    # Sort so there is a fixed order to the resulting tree and create a dict
    # classname : superclassname
    cln_supercln_sorted = NocaseDict()
    for key in sorted(cln_to_supercln.keys()):
        cln_supercln_sorted[key] = cln_to_supercln[key]

    return cln_supercln_sorted


def build_subcln_in_cln(cln_to_supercln):
    """
    Build a dictionary of the direct subclasses for each subclass in
    the cln_to_supercln dictionary
    """

    # Build the class to subclass dictionary from the
    # superclass to class dictionary by reversing the dictionary.
    # Built within a comprehension but comprehension not assigned.
    subcln_in_cln = NocaseDict()
    # pylint: disable=bad-continuation, expression-not-assigned
    [subcln_in_cln.setdefault(v, []).append(k) for (k, v) in
        six.iteritems(cln_to_supercln)]  # noqa: F841

    return subcln_in_cln


def build_nested_tree(class_subclass_dict, top_class_name):
    """
    Build a dictionary structure based on the class/subclass relationships
    in the classname to subclassname tree dictionary provided.

    Returns the dictionary structure in a form suitable for ascii tree
    display

    Parameters:
      class_subclass_dict ():
    """
    def _tree_node(cln_to_subcln_dict, cln):
        """
        Build dictionary of the class/subclass relationships for class cln
        in dictionary of class_subclass names. This maps the input dictionary,
        a flat class_name: [subclass_name] dictionary, to a tree
        of dictionaries of the class names for the hierarchy of classes.

        Returns dictionary of dictionaries in form suitable for asciitree

        Parameters:
          cln_to_subcln_dict:
            Dictionary with all class names as keys and the
            corresponding subclass names for each class name as a list of
            class names

          cln (:term:`string`):
            Class for which a tree node will be generated.  An empty node
            is generated if class not in class_to_subclass_dict.

        Returns:
          Structure of nested dictionaries defining the class/subclass structure
        """
        node_dict = NocaseDict()
        # If there is no subclass, the class will not exist in this dictionary
        if cln in cln_to_subcln_dict:
            cln_list = cln_to_subcln_dict[cln]
            # This should not be necessary if end nodes are not in the dict.
            if cln_list:
                for key in cln_list:
                    node_dict[key] = _tree_node(cln_to_subcln_dict, key)
        return node_dict

    nested_tree_dict = NocaseDict()
    # _tree_node generates dictionary node for elements in class-subclass
    # dictionary and returns complete node structure. This is recursive,
    # with _tree_node recursively calling until there are no subclasses.
    nested_tree_dict[top_class_name] = _tree_node(class_subclass_dict,
                                                  top_class_name)
    return nested_tree_dict



# (C) Copyright 2017 IBM Corp.
# (C) Copyright 2017 Inova Development Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defines click options that are used for multiple scommands and that have
the same definition throughout the environment.  This allows the characteristics
and help to be defined once and used multiple times.
"""

from __future__ import absolute_import, print_function

import click

# Generally we define common options each as a list to encorage using the
# add_options for each option specified in a command definition rather than
# gathering them together in in the @add_options specification. The only
# exception is the qualifier_filters that are ONLY used as a group.

#
# property_list option - Defined here because the option is used in
# multiple places in the command structure.
#
propertylist_option = [                      # pylint: disable=invalid-name
    click.option('--pl', '--propertylist', 'propertylist', multiple=True,
                 type=str,
                 default=None, metavar='PROPERTYLIST',
                 help=u'Filter the properties included in the returned '
                      u'object(s). '
                      u'Multiple properties may be specified with either a '
                      u'comma-separated list or by using the option multiple '
                      u'times. Properties specified in this option that are '
                      u'not in the object(s) will be ignored. '
                      u'The empty string will include no properties. '
                      u'Default: Do not filter properties.')]

names_only_option = [                      # pylint: disable=invalid-name
    click.option('--no', '--names-only', 'names_only', is_flag=True,
                 required=False,
                 help=u'Retrieve only the object paths (names). '
                      u'Default: Retrieve the complete objects including '
                      u'object paths.')]

include_classorigin_instance_option = [         # pylint: disable=invalid-name
    click.option('--ico', '--include-classorigin', 'include_classorigin',
                 is_flag=True, required=False,
                 help=u'Include class origin information in the returned '
                      u'instance(s). '
                      u'Some servers may ignore this option. '
                      u'Default: Do not include class origin information.')]

include_classorigin_class_option = [            # pylint: disable=invalid-name
    click.option('--ico', '--include-classorigin', 'include_classorigin',
                 is_flag=True, required=False,
                 help=u'Include class origin information in the returned '
                      u'class(es). '
                      u'Default: Do not include class origin information.')]

namespace_option = [                     # pylint: disable=invalid-name
    click.option('-n', '--namespace', type=str,
                 required=False, metavar='NAMESPACE',
                 help=u'Namespace to use for this command, instead of the '
                      u'default namespace of the connection.')]

summary_option = [              # pylint: disable=invalid-name
    click.option('-s', '--summary', is_flag=True, required=False,
                 help=u'Show only a summary (count) of the objects.')]

verify_option = [              # pylint: disable=invalid-name
    click.option('-V', '--verify', is_flag=True, required=False,
                 help=u'Prompt for confirmation before performing a change, '
                      u'to allow for verification of parameters. '
                      u'Default: Do not prompt for confirmation.')]

multiple_namespaces_option = [              # pylint: disable=invalid-name
    click.option('-n', '--namespace', type=str, multiple=True,
                 required=False, metavar='NAMESPACE',
                 help=u'Add a namespace to the search scope. '
                      u'May be specified multiple times. '
                      u'Default: Search in all namespaces of the server.')]

#
#  The following options are implement the filtering of class request
#  operations to filter by selected class qualifiers. Note that these
#  options are NOT defined within a list since they are intended to be
#  used only as a list of all of the items
#
# pylint: disable=invalid-name
association_filter_option = \
    click.option('--association/--no-association',
                 default=None,
                 help=u'Filter the returned classes to return only indication '
                      u'classes (--association) or classes that are not '
                      u'associations(--no-association). If the option is not '
                      u'defined no filtering occurs')

indication_filter_option = \
    click.option('--indication/--no-indication',
                 default=None,
                 help=u'Filter the returned classes to return only indication '
                      u'classes (--indication) or classes that are not '
                      u'indications (--no-indication). If the option is not '
                      u'defined no filtering occurs')

experimental_filter_option = \
    click.option('--experimental/--no-experimental',
                 default=None,
                 help=u'Filter the returned classes to return only '
                      u'experimental classes (--experimental) or classes that '
                      u'are not experimental (--no-iexperimental). If the '
                      u'option is not defined no filtering occurs')

deprecated_filter_option = \
    click.option('--deprecated/--no-deprecated',
                 default=None,
                 help=u'Filter the returned classes to return only deprecated '
                      u'classes (--deprecated) or classes that are not '
                      u'deprecated (--no-deprecated). If the option is not '
                      u'defined no filtering occurs')

since_filter_option = \
    click.option('--since',
                 required=False, metavar='VERSION', type=str,
                 default=None,
                 help=u'Filter the returned classes to return only classes  '
                      u'with a version qualifier ge the supplied string. The '
                      u'string must define a version of the form M.N.V '
                      u'consistent the definitions of the VERSION qualifier.')

schema_filter_option = \
    click.option('--schema',
                 required=False, metavar='SCHEMA', type=str,
                 default=None,
                 help=u'Filter the returned classes to return only classes '
                      u'where the classname scheme component (characters '
                      u'before the "_" match the scheme provided.')


subclassof_filter_option = \
    click.option('--subclass-of',
                 required=False, metavar='CLASSNAME', type=str,
                 default=None,
                 help=u'Filter the returned classes to return only classes '
                      u'that are a subclass of the option value.')


leafclass_filter_option = \
    click.option('--leaf-classes', is_flag=True,
                 default=None,
                 help=u'Filter the returned classes to return only leaf '
                      u'(classes; classes with no subclass.')

# pylint: enable=invalid-name

# List of the class filter options that are common to multiple class commands
# Since the filters are in a list to allow them to be used individually, the
# first item of each list must be used for the combined defintion that can
# be use with add_options
# NOTE: Since these options are only used as a group then do not each need to be
# defined as a list above. However, we are trying to define all individual
# common options the same way (in a list).
class_filter_options = [              # pylint: disable=invalid-name
    association_filter_option,
    indication_filter_option,
    experimental_filter_option,
    deprecated_filter_option,
    since_filter_option,
    schema_filter_option,
    subclassof_filter_option,
    leafclass_filter_option
]

help_option = [              # pylint: disable=invalid-name
    click.help_option('-h', '--help', help=u'Show this help message.')]


def add_options(options):
    """
    Accumulate single or multiple options into a list. This list can be
    referenced as a click decorator @add_options(name_of_list)

    The list is reversed because of the way click processes options

    Since this is used with the click decorator add_options and subject to
    future variations of that decorator implementation.

    This is used by pywbemcli to implement common option definitions outside
    of each command definition and include them with the click decorator.

    Parameters:

      options: List of click.option definitions

    Returns:
        the callable _add_options which used by the decorator to do the
        reversal.

    """
    def _add_options(func):
        """
        Reverse options list, else if single item,
        make into a list.
        """
        for option in reversed(options):
            func = option(func)
        return func

    # Return the callable _add_options
    return _add_options
